"""
Demonstrate aXe trace polynomials.

Initial code taken from `(Brammer, Pirzkal, & Ryan 2014) <https://github.com/WFC3Grism/CodeDescription>`_, which contains a detailed
explanation how the grism configuration parameters and coefficients are defined and evaluated.
"""

import os
from collections import OrderedDict

import numpy as np

from . import GRIZLI_PATH, utils


class aXeConf():
    def __init__(self, conf_file='WFC3.IR.G141.V2.5.conf'):
        """Read an aXe-compatible configuration file

        Parameters
        ----------
        conf_file: str
            Filename of the configuration file to read

        """
        if conf_file is not None:
            self.conf = self.read_conf_file(conf_file)
            self.conf_dict = self.conf
            self.conf_file = conf_file
            self.count_beam_orders()

            # Global XOFF/YOFF offsets
            if 'XOFF' in self.conf.keys():
                self.xoff = np.float(self.conf['XOFF'])
            else:
                self.xoff = 0.

            if 'YOFF' in self.conf.keys():
                self.yoff = np.float(self.conf['YOFF'])
            else:
                self.yoff = 0.

    def read_conf_file(self, conf_file='WFC3.IR.G141.V2.5.conf'):
        """Read an aXe config file, convert floats and arrays

        Parameters
        ----------
        conf_file: str
            Filename of the configuration file to read.

        Parameters are stored in an OrderedDict in `self.conf`.
        """
        conf = OrderedDict()
        fp = open(conf_file)
        lines = fp.readlines()
        fp.close()

        for line in lines:
            # empty / commented lines
            if (line.startswith('#')) | (line.strip() == '') | ('"' in line):
                continue

            # split the line, taking out ; and # comments
            spl = line.split(';')[0].split('#')[0].split()
            param = spl[0]
            if len(spl) > 2:
                value = np.cast[float](spl[1:])
            else:
                try:
                    value = float(spl[1])
                except:
                    value = spl[1]

            conf[param] = value

        return conf

    def count_beam_orders(self):
        """Get the maximum polynomial order in DYDX or DLDP for each beam
        """
        self.orders = {}
        for beam in ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J']:
            order = 0
            while 'DYDX_{0:s}_{1:d}'.format(beam, order) in self.conf.keys():
                order += 1

            while 'DLDP_{0:s}_{1:d}'.format(beam, order) in self.conf.keys():
                order += 1

            self.orders[beam] = order-1

    def get_beams(self):
        """Get beam parameters and read sensitivity curves
        """
        import os
        from astropy.table import Table, Column

        self.dxlam = OrderedDict()
        self.nx = OrderedDict()
        self.sens = OrderedDict()
        self.beams = []

        for beam in self.orders:
            if self.orders[beam] > 0:
                self.beams.append(beam)
                self.dxlam[beam] = np.arange(self.conf['BEAM{0}'.format(beam)].min(), self.conf['BEAM{0}'.format(beam)].max(), dtype=int)
                self.nx[beam] = int(self.dxlam[beam].max()-self.dxlam[beam].min())+1
                self.sens[beam] = Table.read('{0}/{1}'.format(os.path.dirname(self.conf_file), self.conf['SENSITIVITY_{0}'.format(beam)]))
                #self.sens[beam].wave = np.cast[np.double](self.sens[beam]['WAVELENGTH'])
                #self.sens[beam].sens = np.cast[np.double](self.sens[beam]['SENSITIVITY'])

                # Need doubles for interpolating functions
                for col in self.sens[beam].colnames:
                    data = np.cast[np.double](self.sens[beam][col])
                    self.sens[beam].remove_column(col)
                    self.sens[beam].add_column(Column(data=data, name=col))

                # Scale BEAM F
                if (beam == 'F') & ('G141' in self.conf_file):
                    self.sens[beam]['SENSITIVITY'] *= 0.35

                if (beam == 'B') & ('G141' in self.conf_file):
                    if self.conf['SENSITIVITY_B'] == 'WFC3.IR.G141.0th.sens.1.fits':
                        self.sens[beam]['SENSITIVITY'] *= 2

                # wave = np.cast[np.double](self.sens[beam]['WAVELENGTH'])
                # sens = np.cast[np.double](self.sens[beam]['SENSITIVITY']
                # self.sens[beam]['WAVELENGTH'] = np.cast[np.double](self.sens[beam]['WAVELENGTH'])
                # self.sens[beam]['SENSITIVITY'] = )

        self.beams.sort()

    def field_dependent(self, xi, yi, coeffs):
        """aXe field-dependent coefficients

        See the `aXe manual <http://axe.stsci.edu/axe/manual/html/node7.html#SECTION00721200000000000000>`_ for a description of how the field-dependent coefficients are specified.

        Parameters
        ----------
        xi, yi : float or array-like
            Coordinate to evaluate the field dependent coefficients, where
            `xi = x-REFX` and `yi = y-REFY`.

        coeffs : array-like
            Field-dependency coefficients

        Returns
        -------
        a : float or array-like
            Evaluated field-dependent coefficients

        """
        # number of coefficients for a given polynomial order
        # 1:1, 2:3, 3:6, 4:10, order:order*(order+1)/2
        if hasattr(coeffs, '__len__'):
            order = int(-1+np.sqrt(1+8*len(coeffs))) // 2
        else:
            order = 1

        # Build polynomial terms array
        # $a = a_0+a_1x_i+a_2y_i+a_3x_i^2+a_4x_iy_i+a_5yi^2+$ ...
        xy = []
        for _p in range(order):
            for _py in range(_p+1):
                # print 'x**%d y**%d' %(p-py, px)
                xy.append(xi**(_p - _py) * yi**(_py))

        # Evaluate the polynomial, allowing for N-dimensional inputs
        a = np.sum((np.array(xy).T*coeffs).T, axis=0)

        return a

    def evaluate_dp(self, dx, dydx):
        """Evalate arc length along the trace given trace polynomial coefficients

        Parameters
        ----------
        dx : array-like
            x pixel to evaluate

        dydx : array-like
            Coefficients of the trace polynomial

        Returns
        -------
        dp : array-like
            Arc length along the trace at position `dx`.

        For `dydx` polynomial orders 0, 1 or 2, integrate analytically.
        Higher orders must be integrated numerically.

        **Constant:**
            .. math:: dp = dx

        **Linear:**
            .. math:: dp = \sqrt{1+\mathrm{DYDX}[1]}\cdot dx

        **Quadratic:**
            .. math:: u = \mathrm{DYDX}[1] + 2\ \mathrm{DYDX}[2]\cdot dx

            .. math:: dp = (u \sqrt{1+u^2} + \mathrm{arcsinh}\ u) / (4\cdot \mathrm{DYDX}[2])

        """
        # dp is the arc length along the trace
        # $\lambda = dldp_0 + dldp_1 dp + dldp_2 dp^2$ ...

        poly_order = len(dydx)-1
        if (poly_order == 2):
            if np.abs(np.unique(dydx[2])).max() == 0:
                poly_order = 1

        if poly_order == 0:  # dy=0
            dp = dx
        elif poly_order == 1:  # constant dy/dx
            dp = np.sqrt(1+dydx[1]**2)*(dx)
        elif poly_order == 2:  # quadratic trace
            u0 = dydx[1]+2*dydx[2]*(0)
            dp0 = (u0*np.sqrt(1+u0**2)+np.arcsinh(u0))/(4*dydx[2])
            u = dydx[1]+2*dydx[2]*(dx)
            dp = (u*np.sqrt(1+u**2)+np.arcsinh(u))/(4*dydx[2])-dp0
        else:
            # high order shape, numerical integration along trace
            # (this can be slow)
            xmin = np.minimum((dx).min(), 0)
            xmax = np.maximum((dx).max(), 0)
            xfull = np.arange(xmin, xmax)
            dyfull = 0
            for i in range(1, poly_order):
                dyfull += i*dydx[i]*(xfull-0.5)**(i-1)

            # Integrate from 0 to dx / -dx
            dpfull = xfull*0.
            lt0 = xfull < 0
            if lt0.sum() > 1:
                dpfull[lt0] = np.cumsum(np.sqrt(1+dyfull[lt0][::-1]**2))[::-1]
                dpfull[lt0] *= -1

            #
            gt0 = xfull > 0
            if gt0.sum() > 0:
                dpfull[gt0] = np.cumsum(np.sqrt(1+dyfull[gt0]**2))

            dp = np.interp(dx, xfull, dpfull)
            if dp[-1] == dp[-2]:
                dp[-1] = dp[-2]+np.diff(dp)[-2]

        return dp

    def get_beam_trace(self, x=507, y=507, dx=0., beam='A', fwcpos=None):
        """Get an aXe beam trace for an input reference pixel and list of output x pixels `dx`

        Parameters
        ----------
        x, y : float or array-like
            Evaluate trace definition at detector coordinates `x` and `y`.

        dx : float or array-like
            Offset in x pixels from `(x,y)` where to compute trace offset and
            effective wavelength

        beam : str
            Beam name (i.e., spectral order) to compute.  By aXe convention,
            `beam='A'` is the first order, 'B' is the zeroth order and
            additional beams are the higher positive and negative orders.

        fwcpos : None or float
            For NIRISS, specify the filter wheel position to compute the
            trace rotation

        Returns
        -------
        dy : float or array-like
            Center of the trace in y pixels offset from `(x,y)` evaluated at
            `dx`.

        lam : float or array-like
            Effective wavelength along the trace evaluated at `dx`.

        """
        NORDER = self.orders[beam]+1

        xi, yi = x-self.xoff, y-self.yoff
        xoff_beam = self.field_dependent(xi, yi, 
                                         self.conf['XOFF_{0}'.format(beam)])
        yoff_beam = self.field_dependent(xi, yi, 
                                         self.conf['YOFF_{0}'.format(beam)])

        # y offset of trace (DYDX)
        dydx = np.zeros(NORDER)  # 0 #+1.e-80
        dydx = [0]*NORDER

        for i in range(NORDER):
            if 'DYDX_{0:s}_{1:d}'.format(beam, i) in self.conf.keys():
                coeffs = self.conf['DYDX_{0:s}_{1:d}'.format(beam, i)]
                dydx[i] = self.field_dependent(xi, yi, coeffs)

        # $dy = dydx_0+dydx_1 dx+dydx_2 dx^2+$ ...

        dy = yoff_beam
        for i in range(NORDER):
            dy += dydx[i]*(dx-xoff_beam)**i

        # wavelength solution
        dldp = np.zeros(NORDER)
        dldp = [0]*NORDER

        for i in range(NORDER):
            if 'DLDP_{0:s}_{1:d}'.format(beam, i) in self.conf.keys():
                coeffs = self.conf['DLDP_{0:s}_{1:d}'.format(beam, i)]
                dldp[i] = self.field_dependent(xi, yi, coeffs)

        self.eval_input = {'x': x, 'y': y, 'beam': beam, 'dx': dx,
                           'fwcpos': fwcpos}
        self.eval_output = {'xi': xi, 'yi': yi, 'dldp': dldp, 'dydx': dydx,
                            'xoff_beam': xoff_beam, 'yoff_beam': yoff_beam,
                            'dy': dy}

        dp = self.evaluate_dp(dx-xoff_beam, dydx)
        # ## dp is the arc length along the trace
        # ## $\lambda = dldp_0 + dldp_1 dp + dldp_2 dp^2$ ...
        # if self.conf['DYDX_ORDER_%s' %(beam)] == 0:   ## dy=0
        #     dp = dx-xoff_beam
        # elif self.conf['DYDX_ORDER_%s' %(beam)] == 1: ## constant dy/dx
        #     dp = np.sqrt(1+dydx[1]**2)*(dx-xoff_beam)
        # elif self.conf['DYDX_ORDER_%s' %(beam)] == 2: ## quadratic trace
        #     u0 = dydx[1]+2*dydx[2]*(0)
        #     dp0 = (u0*np.sqrt(1+u0**2)+np.arcsinh(u0))/(4*dydx[2])
        #     u = dydx[1]+2*dydx[2]*(dx-xoff_beam)
        #     dp = (u*np.sqrt(1+u**2)+np.arcsinh(u))/(4*dydx[2])-dp0
        # else:
        #     ## high order shape, numerical integration along trace
        #     ## (this can be slow)
        #     xmin = np.minimum((dx-xoff_beam).min(), 0)
        #     xmax = np.maximum((dx-xoff_beam).max(), 0)
        #     xfull = np.arange(xmin, xmax)
        #     dyfull = 0
        #     for i in range(1, NORDER):
        #         dyfull += i*dydx[i]*(xfull-0.5)**(i-1)
        #
        #     ## Integrate from 0 to dx / -dx
        #     dpfull = xfull*0.
        #     lt0 = xfull <= 0
        #     if lt0.sum() > 1:
        #         dpfull[lt0] = np.cumsum(np.sqrt(1+dyfull[lt0][::-1]**2))[::-1]
        #         dpfull[lt0] *= -1
        #     #
        #     gt0 = xfull >= 0
        #     if gt0.sum() > 0:
        #         dpfull[gt0] = np.cumsum(np.sqrt(1+dyfull[gt0]**2))
        #
        #     dp = np.interp(dx-xoff_beam, xfull, dpfull)

        # Evaluate dldp
        lam = dp*0.
        for i in range(NORDER):
            lam += dldp[i]*dp**i

        # NIRISS rotation?
        if fwcpos is not None:
            if 'FWCPOS_REF' not in self.conf.keys():
                print('Parameter fwcpos={0} supplied but no FWCPOS_REF in {1:s}'.format(fwcpos, self.conf_file))
                return dy, lam

            order = self.conf['DYDX_ORDER_{0}'.format(beam)]
            if order != 2:
                print('ORDER={0:d} not supported for NIRISS rotation'.format(order))
                return dy, lam

            theta = (fwcpos - self.conf['FWCPOS_REF'])/180*np.pi*1
            theta *= -1  # DMS rotation
            # print('DMS')

            if theta == 0:
                return dy, lam

            # For the convention of swapping/inverting axes for GR150C
            # if 'GR150C' in self.conf_file:
            #     theta = -theta

            # If theta is small, use a small angle approximation.
            # Otherwise, 1./tan(theta) blows up and results in numerical
            # noise.
            xp = (dx-xoff_beam)/np.cos(theta)
            if (1-np.cos(theta) < 5.e-8):
                #print('Approximate!', xoff_beam, np.tan(theta))
                dy = dy + (dx-xoff_beam)*np.tan(theta)
                delta = 0.
                # print('Approx')
            else:
                # Full transformed trace coordinates
                c = dydx
                #print('Not approx')

                beta = c[1]+2*c[2]*xp-1/np.tan(theta)
                chi = c[0]+c[1]*xp+c[2]*xp**2
                if theta < 0:
                    psi = (-beta+np.sqrt(beta**2-4*c[2]*chi))
                    psi *= 1./2/c[2]/np.tan(theta)
                    delta = psi*np.tan(theta)
                    dy = dx*np.tan(theta) + psi/np.cos(theta)
                else:
                    psi = (-beta-np.sqrt(beta**2-4*c[2]*chi))
                    psi *= 1./2/c[2]/np.tan(theta)
                    delta = psi*np.tan(theta)
                    dy = dx*np.tan(theta) + psi/np.cos(theta)

            # Evaluate wavelength at 'prime position along the trace
            dp = self.evaluate_dp(xp+delta, dydx)

            lam = dp*0.
            for i in range(NORDER):
                lam += dldp[i]*dp**i

        return dy, lam

    def show_beams(self, xy=None, beams=['E', 'D', 'C', 'B', 'A']):
        """
        Make a demo plot of the beams of a given configuration file
        """
        import matplotlib.pyplot as plt

        x0, x1 = 507, 507
        dx = np.arange(-800, 1200)

        if 'WFC3.UV' in self.conf_file:
            x0, x1 = 2073, 250
            dx = np.arange(-1200, 1200)
        if 'G800L' in self.conf_file:
            x0, x1 = 2124, 1024
            dx = np.arange(-1200, 1200)
        
        if xy is not None:
            x0, x1 = xy
            
        s = 200  # marker size
        fig, ax = plt.subplots(figsize=[10, 3])
        ax.scatter(0, 0, marker='s', s=s, color='black', edgecolor='0.8',
                    label='Direct')

        for beam in beams:
            if 'XOFF_{0}'.format(beam) not in self.conf.keys():
                continue

            xoff = self.field_dependent(x0, x1, self.conf['XOFF_{0}'.format(beam)])
            dy, lam = self.get_beam_trace(x0, x1, dx=dx, beam=beam)
            xlim = self.conf['BEAM{0}'.format(beam)]
            ok = (dx >= xlim[0]) & (dx <= xlim[1])
            sc = ax.scatter(dx[ok]+xoff, dy[ok], c=lam[ok]/1.e4, marker='s', s=s,
                        alpha=0.5, edgecolor='None')
            ax.text(np.median(dx[ok]), np.median(dy[ok])+1, beam,
                     ha='center', va='center', fontsize=14)
            print('Beam {0}, lambda=({1:.1f} - {2:.1f})'.format(beam, lam[ok].min(), lam[ok].max()))

        ax.grid()
        ax.set_xlabel(r'$\Delta x$' + f' (x0={x0})')
        ax.set_ylabel(r'$\Delta y$' + f' (y0={x1})')

        cb = plt.colorbar(sc, pad=0.01, fraction=0.05)
        cb.set_label(r'$\lambda\,(\mu\mathrm{m})$')
        ax.set_title(self.conf_file)
        fig.tight_layout(pad=0.1)
        #plt.savefig('{0}.pdf'.format(self.conf_file))
        
        return fig
        

def get_config_filename(instrume='WFC3', filter='F140W',
                        grism='G141', module=None, chip=1):
    """Generate a config filename based on the instrument, filter & grism combination.

    Config files assumed to be found the directory specified by the `$GRIZLI`
    environment variable, i.e., `${GRIZLI}/CONF`.

    Parameters
    ----------
    instrume : {'ACS', 'WFC3', 'NIRISS', 'NIRCam', 'WFIRST'}
        Instrument used

    filter : str
        Direct image filter.  This is only used for WFC3/IR, where the grism
        configuration files have been determined for each direct+grism
        combination separately based on the filter wedge offsets of the
        filters.

    grism : str
        Grism name.  Valid combinations are the following:

            ACS : G800L (assumed)
            WFC3 : G102, G141
            NIRISS : GR150R, GR150C
            NIRCam : F322W2, F356W, F430M, F444W, F460M
            WFIRST : (basic assumptions about the WFI grism)

    chip : int
        For ACS/WFC and UVIS, specifies the chip to use.  Note that this
        is switched with respect to the header EXTNAME extensions:

            EXTVER = 1 is extension 1 / (SCI,1) of the flt/flc files but
            corresponds to CCDCHIP = 2 and the ACS.WFC3.CHIP2 config files.

            and

            EXTVER = 2 is extension 4 / (SCI,2) of the flt/flc files but
            corresponds to CCDCHIP = 1 and the ACS.WFC3.CHIP1 config files.

    Returns
    -------
    conf_file : str
        String path of the configuration file.

    """
    if instrume == 'ACS':
        conf_file = os.path.join(GRIZLI_PATH,
                    'CONF/ACS.WFC.CHIP{0:d}.Stars.conf'.format(chip))
        if not os.path.exists(conf_file):
            conf_file = os.path.join(GRIZLI_PATH,
                        'CONF/ACS.WFC.CHIP{0:d}.Cycle13.5.conf'.format(chip))

    if instrume == 'WFC3':
        if grism == 'G280':
            conf_file = os.path.join(GRIZLI_PATH, 'CONF/G280/',
         'WFC3.UVIS.G280.cal/WFC3.UVIS.G280.CHIP{0:d}.V2.0.conf'.format(chip))

            return conf_file

        conf_file = os.path.join(GRIZLI_PATH,
                                 'CONF/{0}.{1}.V4.32.conf'.format(grism, filter))

        # When direct + grism combination not found for WFC3 assume F140W
        if not os.path.exists(conf_file):
            conf_file = os.path.join(GRIZLI_PATH,
                                 'CONF/{0}.{1}.V4.32.conf'.format(grism, 'F140W'))

    if instrume == 'NIRISS':
        
        conf_files = []
        conf_files.append(os.path.join(GRIZLI_PATH,
                            'CONF/{0}.{1}.220725.conf'.format(grism, filter)))
        conf_files.append(os.path.join(GRIZLI_PATH,
                            'CONF/{0}.{1}.conf'.format(grism, filter)))
        conf_files.append(os.path.join(GRIZLI_PATH,
                             'CONF/NIRISS.{0}.conf'.format(filter)))
        
        for conf_file in conf_files:
            if os.path.exists(conf_file):
                #print(f'NIRISS: {conf_file}')
                break
                
        # if not os.path.exists(conf_file):
        #     print('CONF/{0}.{1}.conf'.format(grism, filter))
        #     conf_file = os.path.join(GRIZLI_PATH,
        #                          'CONF/NIRISS.{0}.conf'.format(filter))

    # if instrume == 'NIRCam':
    #     conf_file = os.path.join(GRIZLI_PATH,
    #         'CONF/aXeSIM_NC_2016May/CONF/NIRCam_LWAR_{0}.conf'.format(grism))
    if instrume in ['NIRCAM']:
        #conf_file = os.path.join(GRIZLI_PATH,
        #                         f'CONF/NIRCam.A.{filter}.{grism}.conf')
        
        fi = grism
        gr = filter[-1] # R, C
        # conf_file = os.path.join(GRIZLI_PATH,
        #             f'CONF/GRISM_NIRCAM/gNIRCAM.{fi}.mod{module}.{gr}.conf')
        #
        # conf_file = os.path.join(GRIZLI_PATH,
        #             f'CONF/GRISM_NIRCAM/V2/NIRCAM_{fi}_mod{module}_{gr}.conf')

        conf_file = os.path.join(GRIZLI_PATH,
                    f'CONF/GRISM_NIRCAM/V4/NIRCAM_{fi}_mod{module}_{gr}.conf')
        
    elif instrume == 'NIRCAMA':
        fi = grism
        gr = filter[-1] # R, C
        conf_file = os.path.join(GRIZLI_PATH,
                    f'CONF/GRISM_NIRCAM/gNIRCAM.{fi}.modA.{gr}.conf')

        #conf_file = os.path.join(GRIZLI_PATH,
        #                         f'CONF/NIRCam.B.{filter}.{grism}.conf')

    elif instrume == 'NIRCAMB':
        fi = grism
        gr = filter[-1] # R, C
        conf_file = os.path.join(GRIZLI_PATH,
                    f'CONF/GRISM_NIRCAM/gNIRCAM.{fi}.modB.{gr}.conf')

        #conf_file = os.path.join(GRIZLI_PATH,
        #                         f'CONF/NIRCam.B.{filter}.{grism}.conf')

    if instrume == 'WFIRST':
        conf_file = os.path.join(GRIZLI_PATH, 'CONF/WFIRST.conf')

    if instrume == 'WFI':
        conf_file = os.path.join(GRIZLI_PATH, 'CONF/Roman.G150.conf')

    if instrume == 'SYN':
        conf_file = os.path.join(GRIZLI_PATH, 'CONF/syn.conf')

    # Euclid NISP, config files @
    # http://www.astrodeep.eu/euclid-spectroscopic-simulations/

    if instrume == 'NISP':
        if grism == 'BLUE':
            conf_file = os.path.join(GRIZLI_PATH, 'CONF/Euclid.Gblue.0.conf')
        else:
            conf_file = os.path.join(GRIZLI_PATH, 'CONF/Euclid.Gred.0.conf')

    return conf_file


class JwstDispersionTransform(object):
    """
    Rotate NIRISS and NIRCam coordinates such that slitless dispersion has
    wavelength increasing towards +x.  Also works for HST, but does nothing.
    """
    def __init__(self, instrument='NIRCAM', module='A', grism='R', conf_file=None, header=None):
        
        self.instrument = instrument
        self.module = module
        self.grism = grism
        
        if conf_file is not None:
            self.base = os.path.basename(conf_file.split('.conf')[0])
        else:
            self.base = None
            
        if conf_file is not None:
            if 'NIRISS' in conf_file:
                # NIRISS_F200W_GR150R.conf
                self.instrument = 'NIRISS'
                self.grism = self.base.split('_')[2][-1]
                self.module = 'A'
            elif 'NIRCAM' in conf_file:
                # NIRCAM_F444W_modA_R.conf
                self.instrument = 'NIRCAM'
                self.grism = self.base[-1]
                self.module = self.base[-3]
            else:
                # NIRCAM_F444W_modA_R.conf
                self.instrument = 'HST'
                self.grism = 'G141'
                self.module = 'A'
        elif header is not None:
            if 'INSTRUME' in header:
                self.instrument = header['INSTRUME']
            
            if 'MODULE' in header:
                self.module = module
            
            if self.instrument == 'NIRCAM':
                if 'PUPIL' in header:
                    self.grism = header['PUPIL']
            else:
                if 'FILTER' in header:
                    self.grism = header['FILTER']


    @property
    def array_center(self):
        """
        Center of rotation
        
        Maybe this is 1020 for NIRISS?
        """
        if self.instrument == 'HST':
            return np.array([507, 507])
        else:
            return np.array([1024, 1024])


    @property
    def rotation(self):
        """
        Clockwise rotation (degrees) from detector to wavelength increasing
        towards +x direction
        """
        if self.instrument == 'NIRCAM':
            if self.module == 'A':
                if self.grism in ['R', 'GRISMR']:
                    rotation = 0.
                elif self.grism in ['C', 'GRISMC']:
                    rotation = 90.
                else:
                    raise ValueError(f'NIRCAM {self.grism} must be GRISMR/C')
                    
            elif self.module == 'B':
                if self.grism in ['R', 'GRISMR']:
                    rotation = 180.
                elif self.grism in ['C', 'GRISMC']:
                    rotation = 90.
                else:
                    raise ValueError(f'NIRCAM {self.grism} must be GRISMR/C')
            else:
                raise ValueError(f'NIRCAM {self.module} must be A/B')
                
        elif self.instrument == 'NIRISS':
            if self.grism in ['GR150R','R']:
                rotation = 270.
            elif self.grism in ['GR150C','C']:
                rotation = 180.
            else:
                raise ValueError(f'NIRISS {self.grism} must be GR150R/C')
            
        else:
            # e.g., WFC3, ACS
            rotation = 0.
            
        return rotation


    @property
    def rot90(self):
        """
        Rotations are all multiples of 90 for now, so 
        compute values that can be passed to `numpy.rot90` for rotating
        2D image arrays
        """
        return int(np.round(self.rotation/90))


    @property 
    def trace_axis(self):
        """Which detector axis corresponds to increasing wavelength
        """
        if self.instrument == 'NIRCAM':
            if self.module == 'A':
                if self.grism == 'R':
                    axis = '+x'
                else:
                    axis = '+y'
            else:
                if self.grism == 'R':
                    axis = '-x'
                else:
                    axis = '+y'

        elif self.instrument == 'NIRISS':
            if self.grism == 'R':
                axis = '-y'
            else:
                axis = '-x'
        else:
            # e.g., WFC3, ACS
            axis = '+x'
            
        return axis


    @staticmethod
    def rotate_coordinates(x, y, theta, center):
        """
        Rotate cartesian coordinates ``x`` and ``y`` by angle ``theta`` 
        (radians) about ``center``
        """
        _mat = np.array([[np.cos(theta), -np.sin(theta)],
                         [np.sin(theta), np.cos(theta)]])
        
        x1 = np.atleast_1d(x)
        y1 = np.atleast_1d(y)
        
        return ((np.array([x1, y1]).T-center).dot(_mat)+center).T


    def forward(self, x, y):
        """
        Forward transform, detector to +x
        """
        theta = self.rotation/180*np.pi
        return self.rotate_coordinates(x, y, theta, self.array_center)


    def reverse(self, x, y):
        """
        Reverse transform, +x to detector
        """
        theta = -self.rotation/180*np.pi
        return self.rotate_coordinates(x, y, theta, self.array_center)


class TransformGrismconf(object):
    """
    Transform GRISMCONF-format configuration files to grizli convention of
    wavelength increasing towards +x 

    See https://github.com/npirzkal/GRISMCONF and config files at, e.g., 
    https://github.com/npirzkal/GRISM_NIRCAM.

    """
    def __init__(self, conf_file=''):
        """
        Parameters
        ----------
        conf_file : str
            Configuration filename
        
        """
        import grismconf
        
        self.conf_file = conf_file
        self.conf = grismconf.Config(conf_file)
        
        self.transform = JwstDispersionTransform(conf_file=conf_file)

        self.order_names = {'A':'+1',
                           'B':'0',
                           'C':'+2',
                           'D':'+3',
                           'E':'-1',
                           'F':'+4'}
        
        self.beam_names = {}
        for k in self.order_names:
            self.beam_names[self.order_names[k]] = k

        self.dxlam = OrderedDict()
        self.nx = OrderedDict()
        self.sens = OrderedDict()
        
        self.xoff = 0.0
        self.yoff = 0.0

        self.conf_dict = {}


    @property
    def orders(self):
        """
        GRISMCONF order names, like '+1', '0', '+2', etc.
        """
        return self.conf.orders


    @property
    def beams(self):
        """
        aXe beam names like 'A','B','C', etc.
        """
        beams = [self.beam_names[k] for k in self.orders]
        return beams


    def get_beam_trace(self, x=1024, y=1024, dx=0., beam='A', fwcpos=None):
        """
        Function analogous to `grizli.grismconf.aXeConf.get_beam_trace` but
        that accounts for the different dispersion axes of JWST grisms
        
        Parameters
        ----------
        x, y : float
            Reference position in the rotated frame

        dx : array-like
            Offset in pixels along the trace

        beam : str
            Grism order, translated from +1, 0, +2, +3, -1 = A, B, C, D, E

        fwcpos : float
            NIRISS rotation *(not implemented)*

        Returns
        -------
        dy : float or array-like
            Center of the trace in y pixels offset from `(x,y)` evaluated at
            `dx`.

        lam : float or array-like
            Effective wavelength along the trace evaluated at `dx`.

        """
        x0 = np.squeeze(self.transform.reverse(x, y))
        
        if self.transform.trace_axis == '+x':
            t_func = self.conf.INVDISPX
            trace_func = self.conf.DISPY
            delta = 1*dx
        elif self.transform.trace_axis == '-x':
            t_func = self.conf.INVDISPX
            trace_func = self.conf.DISPY
            delta = -1*dx
        elif self.transform.trace_axis == '+y':
            t_func = self.conf.INVDISPY
            trace_func = self.conf.DISPX
            delta = 1*dx
        else: # -y
            t_func = self.conf.INVDISPY
            trace_func = self.conf.DISPX
            delta = -1*dx

        #print('xref: ', self.conf_file, self.transform.trace_axis)
        #print(x0, t_func)

        t = t_func(self.order_names[beam], *x0, delta)
        tdx = self.conf.DISPX(self.order_names[beam], *x0, t)
        tdy = self.conf.DISPY(self.order_names[beam], *x0, t)

        rev = self.transform.forward(x0[0]+tdx, x0[1]+tdy)
        trace_dy = rev[1,:] - y
        #trace_dy = y - rev[1,:]
        
        # Trace offsets for NIRCam
        if 'V4/NIRCAM_F444W_modB_R.conf' in self.conf_file:
            trace_dy += -0.5
        elif 'V4/NIRCAM_F444W_modA_R.conf' in self.conf_file:
            trace_dy += -2.5
        elif os.path.basename(self.conf_file) == 'NIRCAM_F444W_modA_R.conf':
            trace_dy += -2.5
        elif os.path.basename(self.conf_file) == 'NIRCAM_F444W_modA_C.conf':
            trace_dy += -0.1
            
        wave = self.conf.DISPL(self.order_names[beam], *x0, t)
        if self.transform.instrument != 'HST':
            wave *= 1.e4

        return trace_dy, wave


    def get_beams(self, nt=512):
        """
        Get beam parameters and read sensitivity curves
        
        Parameters
        ----------
        nt : int
            Number of points to sample the GRISMCONF `t` parameter
        
        Returns
        -------
        sets `dxlam`, `nx`, `sens`, attributes
        
        """
        import os
        from astropy.table import Table, Column

        t = np.linspace(0, 1, nt)

        for beam in self.beams:
            order = self.order_names[beam]
            dx = self.conf.DISPX(order, *self.transform.array_center, t)
            dy = self.conf.DISPY(order, *self.transform.array_center, t)
            lam = self.conf.DISPL(order, *self.transform.array_center, t)

            trace_axis = self.transform.trace_axis

            if trace_axis == '+x':
                xarr = 1*dx
            elif trace_axis == '-x':
                xarr = -1*dx
            elif trace_axis == '+y':
                xarr = 1*dy
            elif trace_axis == '-y':
                xarr = -1*dy

            xarr = np.cast[int](np.round(xarr))

            #self.beams.append(beam)
            self.dxlam[beam] = np.arange(xarr[0], xarr[-1], dtype=int)
            self.nx[beam] = xarr[-1] - xarr[0]+1
                        
            sens = Table()
            sens['WAVELENGTH'] = lam.astype(np.double)
            sens['SENSITIVITY'] = self.conf.SENS[order](lam).astype(np.double)
            if lam.max() < 100:
                sens['WAVELENGTH'] *= 1.e4

            self.sens[beam] = sens
            
            self.conf_dict[f'BEAM{beam}'] = np.array([xarr[0], xarr[-1]])
            self.conf_dict[f'MMAG_EXTRACT_{beam}'] = 29


def load_grism_config(conf_file, warnings=True):
    """Load parameters from an aXe configuration file

    Parameters
    ----------
    conf_file : str
        Filename of the configuration file

    Returns
    -------
    conf : `~grizli.grismconf.aXeConf`
        Configuration file object.  Runs `conf.get_beams()` to read the
        sensitivity curves.
    """
    if 'V3/NIRCAM' in conf_file:
        conf = TransformGrismconf(conf_file)
        conf.get_beams()
    elif 'V2/NIRCAM' in conf_file:
        conf = TransformGrismconf(conf_file)
        conf.get_beams()
    elif 'V4/NIRCAM' in conf_file:
        conf = TransformGrismconf(conf_file)
        conf.get_beams()
    else:
        conf = aXeConf(conf_file)
        conf.get_beams()
    
    # Preliminary hacks for on-sky NIRISS
    if 'GR150' in conf_file:
        if 0:
            hack_niriss = 1./1.8 * 1.1
        
            msg = f"""
     ! Scale NIRISS sensitivity by {hack_niriss:.3f} to hack gain correction
     ! and match GLASS MIRAGE simulations. Sensitivity will be updated when
     ! on-sky data available
     """
            msg = f' ! Scale NIRISS sensitivity by {hack_niriss:.3f} prelim flux correction'
            utils.log_comment(utils.LOGFILE, msg, verbose=warnings)
        else:
            hack_niriss = 1.0
            
        for b in conf.sens:
            conf.sens[b]['SENSITIVITY'] *= hack_niriss
            if 'ERROR' in conf.sens[b].colnames:
                conf.sens[b]['ERROR'] *= hack_niriss
        
        if 'F115W' in conf_file:
            pass
            # msg = f""" !! Shift F115W along dispersion"""
            # utils.log_comment(utils.LOGFILE, msg, verbose=warnings)
            # for b in conf.beams:
            #     #conf.conf[f'DYDX_{b}_0'][0] += 0.25
            #     conf.conf[f'DLDP_{b}_0'] -= conf.conf[f'DLDP_{b}_1']*0.5
        else:
            msg = f""" !! Shift {os.path.basename(conf_file)} along dispersion"""
            utils.log_comment(utils.LOGFILE, msg, verbose=warnings)
            for b in conf.beams:                    
                #conf.conf[f'DYDX_{b}_0'][0] += 0.25
                conf.conf[f'DLDP_{b}_0'] += conf.conf[f'DLDP_{b}_1']*0.5
                
                # For red galaxy
                conf.conf[f'DLDP_{b}_0'] += conf.conf[f'DLDP_{b}_1']*0.5                
                # if 'F200W' in conf_file:
                #     conf.conf[f'DLDP_{b}_0'] += conf.conf[f'DLDP_{b}_1']*0.5

                
        #     _w = conf.sens['A']['WAVELENGTH']
        #     _w0 = (_w*conf.sens['A']['SENSITIVITY']).sum()
        #     _w0 /=  conf.sens['A']['SENSITIVITY'].sum()
        #     slope = 1.05 + 0.2 * (_w - _w0)/3000
        #     # print('xxx', conf_file, _w0)
        #     conf.sens['A']['SENSITIVITY'] *= slope
        
        if ('F150W' in conf_file) & (hack_niriss > 1.01):
            conf.sens['A']['SENSITIVITY'] *= 1.08
            
        # Scale 0th orders in F150W,F200W
        if ('F150W' in conf_file) | ('F200W' in conf_file):
            msg = f""" ! Scale 0th order (B) by an additional x 1.5"""
            utils.log_comment(utils.LOGFILE, msg, verbose=warnings)
            conf.sens['B']['SENSITIVITY'] *= 1.5
            if 'ERROR' in conf.sens['B'].colnames:
                conf.sens['B']['ERROR'] *= 1.5
        
        # Another shift from 0723, 2744
        # if ('GR150C.F200W' in conf_file):
        #     msg = f""" !! Extra shift for GR150C.F200W"""
        #     utils.log_comment(utils.LOGFILE, msg, verbose=warnings)
        #     for b in conf.beams:
        #         pass
        #         #conf.conf[f'DYDX_{b}_0'][0] += 0.25
        #         # conf.conf[f'DLDP_{b}_0'] -= conf.conf[f'DLDP_{b}_1']*0.5
        
        # Shift x by 1 px
        # msg = f""" ! Shift NIRISS by 0.5 pix along dispersion direction"""
        # utils.log_comment(utils.LOGFILE, msg, verbose=warnings)
        # 
        # for b in conf.beams:
        #     conf.conf[f'DLDP_{b}_0'] -= conf.conf[f'DLDP_{b}_1']*0.5
                
    return conf



"""
Splitting fields into tiles
"""
import os
import glob

import numpy as np
import astropy.wcs as pywcs
from collections import OrderedDict
import yaml

from .. import utils

def make_all_fields():
    """
    Define tile layout
    """
    
    layout = """# field ra dec dx dy
    abell370   39.9750000 -1.5752778   40 40 
    macs0416   64.0333333 -24.0730556  40 40  
    macs1149   177.4000000 22.3972222  40 40  
    abell2744   3.5875000 -30.3966667  40 40
    macs0417   64.3958333 -11.9091667  40 40 
    macs1423   215.9500000 24.0811111  40 40 
    sgas1723    260.895 34.205 20 20
    sgas1226    186.69  21.88  10 10
    dracoii    238.1983333 64.56528 20 20
    whl0137    24.355  -8.457 22 22
    abell1689   197.8765  -1.328 22 22
    cos      150.125 2.2 90 90
    j013804m2156   24.51616  -21.93046 40 40 
    j015020m1006   27.58366  -10.09244 40 40 
    j045412m0302   73.547   -3.013     40 40 
    j091840m8103  139.64290  -81.05227 40 40 
    j134200m2442  205.50969  -24.69773 40 40 
    j152256p2536  230.72345   25.59692 40 40 
    j212928m0741  322.358   -7.69209 40 40 
    gds 53.1592464 -27.7828223 46 46
    egs 214.8288    52.8067333 60 60
    gdn 189.2369029 62.2312839 46 46
    uds 34.40869 -5.16299 30 30
    smacs0723 110.83403 -73.45429 20 20
    j1235 189.025 4.948 20 20
    """
    
    tile_defs = utils.read_catalog(layout)
    
    
def define_tiles(ra=109.3935148, dec=37.74934031, size=(24, 24), tile_size=6, overlap=0.3, field='macs0717', pixscale=0.05, theta=0):
    """
    Tile definition
    """
    
    tile_wcs = OrderedDict()
    
    size_per = tile_size-overlap
    nx = int(np.ceil(size[0]/size_per))
    ny = int(np.ceil(size[1]/size_per))
    
    sx = nx*tile_size-(nx-1)*overlap
    sy = ny*tile_size-(ny-1)*overlap
    px = int(tile_size*60/pixscale)

    # Even number of pixels
    header, parent_wcs = utils.make_wcsheader(ra=ra, dec=dec,
                                              size=(sx*60, sy*60), 
                                              pixscale=pixscale, theta=theta)
    for i in range(nx):
        x0 = int(i*px-i*(overlap/tile_size*px))
        slx = slice(x0, x0+px)
        for j in range(ny):
            y0 = int(j*px-j*(overlap/tile_size*px))
            sly = slice(y0, y0+px)
            
            slice_header = utils.get_wcs_slice_header(parent_wcs, slx, sly)
            slice_wcs = pywcs.WCS(slice_header)
            slice_wcs._header = slice_header
            #tile=(i+1)*10+(j+1)
            tile = '{0:02d}.{1:02d}'.format(i+1, j+1)
            tile_wcs[tile] = slice_wcs
    
    tile_dict = {}
    for k in tile_wcs:
        _h = utils.to_header(tile_wcs[k])
        tile_dict[k] = {'field':field}
        for key in _h:
            tile_dict[k][key] = _h[key]
    
    yml_file = '{0}_{1:02d}mas_tile_wcs.yaml'.format(field, int(pixscale*1000))
    with open(yml_file,'w') as fp:
        yaml.dump(tile_dict, stream=fp)
                                                          
    #np.save(,[tile_wcs])
    reg_file = yml_file.replace('.yaml', '.reg')
    with open(reg_file,'w') as fpr:
        fpr.write('fk5\n')
        for t in tile_wcs:
            fp = tile_wcs[t].calc_footprint()
            pstr = 'polygon('
            pstr += ','.join(['{0:.6f}'.format(i)
                            for i in fp.flatten()])
            pstr += ') # text={{{0}}}\n'.format(t)
            fpr.write(pstr)
    
    return tile_wcs


def make_field_tiles(field='cos', ra=150.125, dec=2.2, rsize=45, tile_npix=2048+256, pscale=0.08, initial_status=90, send_to_database=False, **kwargs):
    """
    Make tiles across COSMOS with a single tangent point
    """
    import numpy as np    
    
    from . import db, tile_mosaic
        
    np.set_printoptions(precision=6)
    
    #field='cos'
    
    # tile_npix = 4096
    # pscale = 0.1
    # 
    # ra, dec = 150.1, 2.29
    # rsize = 48
    # overlap = 0
    # 
    # # New tiles with overlaps
    # ra, dec = 150.125, 2.2
    # rsize = 45
    # pscale = 0.080
    # tile_npix = 2048+256
    
    if 0:
        ra, dec, rsize = 34.40869, -5.16299, 30
        field = 'uds'
        pscale = 0.08
        tile_npix = 2048+256
        
        
    tile_arcmin = tile_npix*pscale/60
    overlap=tile_arcmin/9
    
    print(f'{field}: tile size, arcmin = {tile_arcmin}')
    
    tiles = define_tiles(ra=ra, dec=dec, 
                         size=(rsize*2, rsize*2), 
                         tile_size=tile_arcmin, 
                         overlap=overlap, field=field, 
                         pixscale=pscale, theta=0)
        
    rows = []
    for t in tiles:
        wcs = tiles[t]
        sr = utils.SRegion(wcs.calc_footprint())
                    
        wcsh = utils.to_header(wcs)
        row = [t]
        for k in wcsh:
            row.append(wcsh[k])
        
        try:
            fp = sr.polystr()[0]
        except:
            fp = sr.xy[0].__str__().replace('\n', ',').replace('   ', ',')
            fp = fp.replace(' -',',-')
            fp = fp.replace(' ', '').replace('[','(')
            fp = fp.replace(']', ')').replace(',,',',')

        row.append(fp)     
        
        rows.append(row)
    
    names = ['tile']
    for k in wcsh:
        names.append(k.lower())
    
    names.append('footprint')

    tiles = utils.GTable(rows=rows, names=names)
    tiles['status'] = initial_status
    
    tiles['field'] = field
            
    #db.send_to_database('cosmos_tiles', cos_tile, if_exists='replace')
    if send_to_database:
        db.execute(f"delete from combined_tiles where field = '{field}'")
        db.send_to_database('combined_tiles', tiles, if_exists='append')
    
    return tiles
    
def split_tiles(root='abell2744-080-08.08', ref_tile=(8,8), filters=['visr','f125w','h'], optical=False, suffix='.rgb', xsize=6, zoom_levels=[4,3,2,1], force=False, scl=1, invert=False, verbose=True, rgb_scl=[1,1,1], rgb_min=-0.01):
    """
    Split image into 256 pixel tiles for map display
    """
    import matplotlib.pyplot as plt
    from skimage.io import imsave
    
    from grizli.pipeline import auto_script
    
    nx = (2048+256)
    dpi = int(nx/xsize)
    
    if os.path.exists(f'{root}{suffix}.png') & (~force):
        return True
    
    try:
        _ = auto_script.field_rgb(root=root,
                                  xsize=xsize, filters=filters, 
                                  full_dimensions=2**optical, HOME_PATH=None, 
                                  add_labels=False,
                                  gzext='*', suffix=suffix, 
                                  output_format='png',
                                  scl=scl, invert=invert, 
                                  rgb_scl=rgb_scl, rgb_min=rgb_min)
    except IndexError:
        return False
    
    fig = _[-1]
    
    base = '_'.join(root.split('-')[:-1]).replace('+','_') + '_' + suffix[1:]
    
    tx, ty = np.cast[int](root.split('-')[-1].split('.'))

    for iz, zoom in enumerate(zoom_levels):
        if iz > 0:
            zoom_img = f'{root}{suffix}.{2**iz:d}.png'
            fig.savefig(zoom_img, dpi=dpi/2**iz)
            img = plt.imread(zoom_img)
        else:
            img = plt.imread(f'{root}{suffix}.png')
        
        if verbose:
            print(f'zoom: {zoom} {img.shape}')
        
        img = img[::-1,:,:]
        
        ntile = int(2048/2**(4-zoom)/256)
        left = (tx - ref_tile[0])*ntile
        bot = -(ty - ref_tile[1])*ntile+2*ntile
        # print(zoom, ntile, left, bot)

        #axes[iz].set_xlim(-ntile*0.1, ntile*(1.1)-1)
        #axes[iz].set_ylim(*axes[iz].get_xlim())

        for i in range(ntile):
            xi = left + i
            for j in range(ntile):
                yi = bot - j - 1
                
                slx = slice((i*256), (i+1)*256)
                sly = slice((j*256), (j+1)*256)
                
                tile_file = f'{root}-tiles/{base}/{zoom}/{yi}/{xi}.png'
                if verbose > 1:
                    print(f'  {i} {j} {tile_file}')
                
                dirs = tile_file.split('/')
                for di in range(1,5):
                    dpath = '/'.join(dirs[:di])
                    #print(dpath)
                    if not os.path.exists(dpath):
                        os.mkdir(dpath)
                                     
                imsave(tile_file, img[sly, slx, :][::-1,:,:],
                       plugin='pil', format_str='png')


def make_all_tile_images(root, force=False, ref_tile=(8,8), cleanup=True, zoom_levels=[4,3,2,1], brgb_filts=['visr','visb','uv'], rgb_filts=['visr','j','h'], blue_is_opt=True, make_opt_filters=True, make_ir_filters=True):
    """
    Make tiles for map
    """
    import matplotlib.pyplot as plt
    
    from grizli.pipeline import auto_script
    
    #root = f'{field}-080-08.08'

    files = glob.glob(f'{root}-[hvu]*')
    files += glob.glob(f'{root}*.rgb.png')

    if len(files) == 0:
        auto_script.make_filter_combinations(root, 
                          filter_combinations={'h':['F140W','F160W'], 
                                               'j':['F105W','F110W','F125W'],
                                            'visr':['F850LP','F814W','F775W'],
                                     'visb':['F606W','F555W', 'F606WU'][:-1],
                                'uv':['F438WU','F435W','F435WU', 'F475W'], 
                                'sw':['F070W-CLEAR','F090W-CLEAR',
                                      'F115W-CLEAR','F150W-CLEAR',
                                      'F200W-CLEAR'], 
                                'lw':['F277W-CLEAR','F356W-CLEAR',
                                      'F410M-CLEAR','F444W-CLEAR'],
                            }, 
                                            weight_fnu=False)

    if 'j' in rgb_filts:
        rgb_scl = [1.1, 0.8, 1]
    else:
        rgb_scl = [1,1,1]
        
    split_tiles(root, ref_tile=ref_tile, 
                filters=rgb_filts, zoom_levels=zoom_levels,
                optical=False, suffix='.rgb', xsize=6, scl=1,
                force=force, rgb_scl=rgb_scl)

    plt.close('all')
    
    if len(glob.glob(f'{root}*.brgb.png')) == 0:
        split_tiles(root, ref_tile=ref_tile, 
                    filters=brgb_filts, zoom_levels=zoom_levels,
                    optical=blue_is_opt, suffix='.brgb', xsize=6, scl=2,
                    force=force, rgb_scl=[1., 1.2, 1.4], rgb_min=-0.018)

        plt.close('all')
    
    # JWST SW
    if len(glob.glob(f'{root}*.swrgb.png')) == 0:
        split_tiles(root, ref_tile=ref_tile, 
                    filters=[f.lower() for f in ['F070W-CLEAR','F090W-CLEAR',
                          'F115W-CLEAR','F150W-CLEAR',
                          'F200W-CLEAR']],
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.swrgb', xsize=6, scl=4,
                    force=force, rgb_scl=[1,1,1], rgb_min=-0.018)

        plt.close('all')
    
    # JWST LW
    if len(glob.glob(f'{root}*.lwrgb.png')) == 0:
        split_tiles(root, ref_tile=ref_tile, 
                    filters=[f.lower() for f in ['F277W-CLEAR','F356W-CLEAR',
                          'F360M-CLEAR','F410M-CLEAR','F444W-CLEAR',
                          'F480M-CLEAR']],
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.lwrgb', xsize=6, scl=4,
                    force=force, rgb_scl=[1,1,1], rgb_min=-0.018)

        plt.close('all')
    
    # All NIRCam
    if len(glob.glob(f'{root}*.ncrgb.png')) == 0:
        split_tiles(root, ref_tile=ref_tile, 
                    filters=[f.lower() for f in ['F070W-CLEAR','F090W-CLEAR',
                                                 'F115W-CLEAR','F150W-CLEAR',
                                                 'F200W-CLEAR',
                                                 'F277W-CLEAR','F356W-CLEAR',
                                                'F410M-CLEAR','F444W-CLEAR']],
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.ncrgb', xsize=6, scl=4,
                    force=force, rgb_scl=[1,1,1], rgb_min=-0.018)

        plt.close('all')
    
    if 'dracoii' in root:
        
        split_tiles(root, ref_tile=ref_tile, 
                    filters=[f.lower() for f in ['F090W-CLEAR',
                             'F150W-CLEAR']],
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.swrgb', xsize=6, scl=4,
                    force=force, rgb_scl=[1,1,1], rgb_min=-0.018)

        plt.close('all')

        split_tiles(root, ref_tile=ref_tile, 
                    filters=[f.lower() for f in ['F360M-CLEAR',
                             'F480M-CLEAR']],
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.lwrgb', xsize=6, scl=4,
                    force=force, rgb_scl=[1,1,1], rgb_min=-0.018)

        plt.close('all')

        split_tiles(root, ref_tile=ref_tile, 
                    filters=[f.lower() for f in ['F090W-CLEAR','F150W-CLEAR',
                                                 'F480M-CLEAR',
                                                ]],
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.ncrgb', xsize=6, scl=4,
                    force=force, rgb_scl=[1.4, 0.6, 0.35], rgb_min=-0.018)

        plt.close('all')
        
    if root.startswith('cos'):
        if len(glob.glob(f'{root}*.vi.png')) == 0:
            split_tiles(root, ref_tile=ref_tile, 
                        filters=['f814w','f160w'], zoom_levels=zoom_levels,
                        optical=False, suffix='.vi', xsize=6, scl=0.8,
                        force=force, rgb_scl=[1, 1, 1], rgb_min=-0.018)

            plt.close('all')
        
    # IR
    if make_ir_filters:
        files = []
        for ir_filt in ['f098m','f105w','f110w','f125w','f140w','f160w']:
            files += glob.glob(f'{root}-{ir_filt}*sci.fits*')
        
        # MIRI on 80mas 
        for ir_filt in ['f560w','f770w','f1000w','f1130w',
                        'f1280w','f1500w','f1800w','f2100w','f2550w']:
            files += glob.glob(f'{root}-{ir_filt}*sci.fits*')
        
        files.sort()
        filts = [('-'.join(file.split(f'{root}-')[1:])).split('_')[0]
                 for file in files]
                 
        for filt in filts:
            if os.path.exists(f'{root}.{filt}.png'):
                continue

            split_tiles(root, ref_tile=ref_tile, 
                    filters=[filt], zoom_levels=zoom_levels,
                    optical=False, suffix=f'.{filt}', xsize=6, 
                    force=force, scl=2, invert=True)

            plt.close('all')
    
    if make_opt_filters:
        # Optical, 2X pix
        files = glob.glob(f'{root}-f[2-8]*sci.fits*')
        files += glob.glob(f'{root}-f*clear*sci.fits*')
        files.sort()
        filts = [file.split(f'{root}-')[1].split('_')[0] for file in files]
        for filt in filts:
            if os.path.exists(f'{root}.{filt}.png'):
                continue

            split_tiles(root, ref_tile=ref_tile, 
                filters=[filt], zoom_levels=zoom_levels,
                optical=True, suffix=f'.{filt}', xsize=6, 
                force=force, scl=2, invert=True)

            plt.close('all')
    
    if cleanup:
        files = glob.glob(f'{root}-[vhuj]*fits*')
        files.sort()
        for file in files:
            print(f'rm {file}')
            os.remove(file)


TILE_FILTERS = ['F336W', 'F350LP', 'F390W', 'F435W', 'F438W', 'F475W',
                'F555W', 'F606W', 'F625W', 'F775W', 
                'F814W', 'F850LP', 
                'F098M', 'F105W', 'F110W', 'F125W', 'F140W', 'F160W',
                'F070W-CLEAR','F090W-CLEAR','F115W-CLEAR','F150W-CLEAR', # NRC
                'F200W-CLEAR','F210M-CLEAR','F250M-CLEAR',
                'F150W2-CLEAR','F322W2-CLEAR',
                'F277W-CLEAR','F356W-CLEAR','F444W-CLEAR',
                'F300M-CLEAR','F360M-CLEAR','F410M-CLEAR', 'F480M-CLEAR',
                'CLEAR-F115W','CLEAR-F150W','CLEAR-F200W', # NIS
                'F560W','F770W','F1000W','F1280W','F1500W', # MIR
                'F1800W','F2100W','F2550W',
                ][1:]


def process_tile(field='cos', tile='01.01', filters=TILE_FILTERS, fetch_existing=True, cleanup=True, make_catalog=False, clean_flt=True, pixfrac=0.75, make_tile_images=True):
    """
    """
    import numpy as np
    
    import astropy.io.fits as pyfits
    import astropy.wcs as pywcs
    from grizli.aws import visit_processor, db
    import golfir.catalog
    from grizli.pipeline import auto_script
    
    utils.LOGFILE = 'mosaic.log'
    
    print(f'Set status=1 in `combined_tiles` for {field} {tile}')
    db.execute(f"update combined_tiles set status=1 where tile = '{tile}' and field='{field}'")
    
    row = db.SQL(f"select * from combined_tiles where tile = '{tile}' and field='{field}'")
    
    h = pyfits.Header()
    for k in row.colnames:
        if k in ['footprint']:
            continue
        
        h[k.upper()] = row[k][0]
        
    ir_wcs = pywcs.WCS(h)
    
    root = f'{field}-080-{tile}'
    
    if fetch_existing:
        os.system(f"""aws s3 sync s3://grizli-v2/ClusterTiles/{field}/ ./
                          --exclude "*"
                          --include "{root}*_dr?*fits.gz"
                          """.replace('\n', ' '))
        
        files = glob.glob(f'{root}*gz')
        for file in files:
            os.system(f'gunzip --force {file}')
            
    visit_processor.cutout_mosaic(rootname=root, 
                              skip_existing=True,
                              ir_wcs=ir_wcs,
                              filters=filters, 
                              pixfrac=pixfrac,
                              kernel='square', s3output=None, 
                              gzip_output=False, clean_flt=clean_flt)
    
    files = glob.glob(f'{root}*dr*fits')
    if len(files) == 0:
        db.execute(f"update combined_tiles set status=4 where tile = '{tile}'")
        print(f'No images found, set status=4 in `combined_tiles` for {field} {tile}')
        
        return True
    
    if make_catalog:
        
        golfir.catalog.make_charge_detection(root, ext='ir')
    
        phot = auto_script.multiband_catalog(field_root=root) #, **phot_kwargs)
    
        if len(phot) == 0:
            # Empty catalog
            db.execute(f"update combined_tiles set status=10 where tile = '{tile}' AND field = '{field}'")
        
            if cleanup:

                files = glob.glob(f'{root}*')
                files.sort()
                for file in files:
                    print(f'rm {file}')
                    os.remove(file)
        
            return False
        
        for i in [4,5,6]:
            for c in phot.colnames:
                if c.endswith(f'{i}'):
                    phot.remove_column(c)
    
        for c in phot.colnames:
            if c in ['ra','dec','x_world','y_world']:
                continue
            
            if phot[c].dtype == np.float64:
                phot[c] = phot[c].astype(np.float32)
            elif phot[c].dtype == np.int64:
                phot[c] = phot[c].astype(np.int32)
              
        phot['tile'] = tile
        phot['field'] = field
    
        if 'id' in phot.colnames:
            phot.remove_column('id')
    
        for c in ['xmin','xmax','ymin','ymax']:
            if c in phot.colnames:
                phot.rename_column(c, c+'pix')
    
        for c in phot.colnames:
            if '-' in c:
                phot.rename_column(c, c.replace('-','_'))
            
        if 0:
            db.execute('CREATE TABLE combined_tile_phot AS SELECT * FROM cosmos_tile_phot limit 0')
            db.execute('ALTER TABLE combined_tile_phot DROP COLUMN id;')
            db.execute('ALTER TABLE combined_tile_phot ADD COLUMN id SERIAL PRIMARY KEY;')
        
        db.execute(f"DELETE from combined_tile_phot WHERE tile='{tile}' and field='{field}'")
    
        seg = pyfits.open(f'{root}-ir_seg.fits')
    
        ### IDs on edge
        edge = np.unique(seg[0].data[16:19,:])
        edge = np.append(edge, np.unique(seg[0].data[-19:-16,:]))
        edge = np.append(edge, np.unique(seg[0].data[:, 16:19]))
        edge = np.append(edge, np.unique(seg[0].data[:, -19:-16]))
        edge = np.unique(edge)
        phot['edge'] = np.in1d(phot['number'], edge)*1
    
        ### Add missing columns
        cols = db.SQL('select * from combined_tile_phot limit 2').colnames
        for c in phot.colnames:
            if c not in cols:
                print('Add column {0} to `combined_tile_phot` table'.format(c))
                if phot[c].dtype in [np.float64, np.float32]:
                    SQL = "ALTER TABLE combined_tile_phot ADD COLUMN {0} real;".format(c)
                else:
                    SQL = "ALTER TABLE combined_tile_phot ADD COLUMN {0} int;".format(c)
                
                db.execute(SQL)
            
        db.send_to_database('combined_tile_phot', phot, if_exists='append')
    
        if 'id' not in cols:
            # Add unique id index column
            db.execute('ALTER TABLE combined_tile_phot ADD COLUMN id SERIAL PRIMARY KEY;')
    
        # Use db id
        ids = db.SQL(f"SELECT number, id, ra, dec, tile from combined_tile_phot WHERE tile='{tile}' AND field='{field}'")
        idx, dr = ids.match_to_catalog_sky(phot)
        
        phot['id'] = phot['number']
        for c in ['xmin','xmax','ymin','ymax']:
            if c+'pix' in phot.colnames:
                phot.rename_column(c+'pix', c)
    
        golfir.catalog.switch_segments(seg[0].data, phot, ids['id'][idx])
        pyfits.writeto(f'{root}-ir_seg.fits', data=seg[0].data, 
                       header=seg[0].header, overwrite=True)
    
    if make_tile_images:
        ### Make subtiles
        ref_tiles = {'cos': (16,16), 
                     'uds': (11, 10),
                     'abell2744': (8, 8), 
                     'egs': (10,14),
                     'gds': (9,9), 
                     'gdn': (9,9),
                     'sgas1723': (4,4),
                     'sgas1226': (2,2),
                     'smacs0723': (4,4), 
                     'dracoii':(5,4),
                     'egs-v2': (10,14),
                     'j1235':(6,4),
                     'whl0137':(5,5),
                     }
    
        if field in ref_tiles:
            ref_tile = ref_tiles[field]
        else:
            ref_tile = (9, 9)
        
        make_all_tile_images(root, force=False, ref_tile=ref_tile,
                             rgb_filts=['h','j','visr'],
                             brgb_filts=['visr','visb','uv'],
                             blue_is_opt=(field not in ['j013804m2156']), 
                             make_ir_filters=True, 
                             make_opt_filters=True)
    
    print(f'Sync ./{root}-tiles/ >> s3://grizli-v2/ClusterTiles/Map/{field}/')
    
    os.system(f'aws s3 sync ./{root}-tiles/ ' + 
              f' s3://grizli-v2/ClusterTiles/Map/{field}/ ' + 
               '--acl public-read --quiet')
                      
    ### Gzip products
    drz_files = glob.glob(f'{root}-*_dr*fits')
    drz_files += glob.glob(f'{root}*seg.fits')
    drz_files.sort()
        
    for file in drz_files:
        cmd = f'gzip --force {file}'
        print(cmd)
        os.system(cmd)
        
    os.system(f'aws s3 sync ./ s3://grizli-v2/ClusterTiles/{field}/' + 
              f' --exclude "*" --include "{root}*gz"' +
              f' --include "{root}*wcs.csv"' +
              f' --include "{root}*fp.png" --acl public-read')
    
    db.execute(f"update combined_tiles set status=2 where tile = '{tile}' AND field = '{field}'")
    print(f'Set status=2 in `combined_tiles` for {field} {tile}')
    
    if cleanup:
        print(f'rm -rf {root}-tiles')
        os.system(f'rm -rf {root}-tiles')
        
        files = glob.glob(f'{root}*')
        files.sort()
        for file in files:
            print(f'rm {file}')
            os.remove(file)
        



def get_random_tile():
    """
    Find a visit that needs processing
    """
    from grizli.aws import db
    
    all_tiles = db.SQL(f"""SELECT tile, field 
                           FROM combined_tiles
                           WHERE status=0""")
    
    if len(all_tiles) == 0:
        return None, None
    
    tile, field = all_tiles[np.random.randint(0, len(all_tiles))]
    return tile, field


def run_one(own_directory=True, **kwargs):
    """
    Run a single random visit
    """
    import os
    import time
    from grizli.aws import db

    ntile = db.SQL("""SELECT count(status)
                       FROM combined_tiles
                       WHERE status = 0""")['count'][0] 
    
    tile, field = get_random_tile()
    if tile is None:
        with open('/GrizliImaging/tile_finished.txt','w') as fp:
            fp.write(time.ctime() + '\n')
    else:
        print(f'============  Run tile  ==============')
        print(f'{field} {tile}')
        print(f'========= {time.ctime()} ==========')
        
        with open('/GrizliImaging/tile_history.txt','a') as fp:
            fp.write(f'{time.ctime()} {tile}\n')
        
        if own_directory:
            path = f'/GrizliImaging/{field}-{tile}'
            if not os.path.exists(path):
                os.makedirs(path)
                
            os.chdir(path)
            
        #process_visit(tile, clean=clean, sync=sync)
        process_tile(tile=tile, field=field, **kwargs)


def create_mosaic_from_tiles(assoc, filt='ir', clean=True):
    """
    Get tiles overlapping visit footprint
    """
    import glob    
    import astropy.io.fits as pyfits
    import astropy.table
    from grizli.aws import db
    
    olap_tiles = db.SQL(f"""SELECT tile, field
                        FROM combined_tiles t, exposure_files e
                        WHERE e.assoc = '{assoc}'
                        AND polygon(e.footprint) && polygon(t.footprint)
                        GROUP BY tile, field""")
    
    tx = np.array([int(t.split('.')[0]) for t in olap_tiles['tile']])
    ty = np.array([int(t.split('.')[1]) for t in olap_tiles['tile']])
    
    xm = tx.min()
    ym = ty.min()
    
    nx = tx.max()-tx.min()+1
    ny = ty.max()-ty.min()+1
    
    print(olap_tiles)
    
    field = olap_tiles['field'][0]
    for t in olap_tiles['tile']:
        # print(f'Fetch tile s3://grizli-v2/ClusterTiles/{field}/{field}-080-{t}-{filt}*')
        # 
        # os.system(f"aws s3 cp s3://grizli-v2/ClusterTiles/{field}/{field}-080-{t}-{filt}_drz_sci.fits.gz . ")
        # os.system(f"aws s3 cp s3://grizli-v2/ClusterTiles/{field}/{field}-080-{t}-{filt}_drz_wht.fits.gz . ")
        # os.system(f"aws s3 cp s3://grizli-v2/ClusterTiles/{field}/{field}-080-{t}-{filt}_seg.fits.gz . ")
        
        os.system(f"""aws s3 sync s3://grizli-v2/ClusterTiles/{field}/ ./
                          --exclude "*"
                          --include "{field}*{t}-{filt}*_dr?*fits.gz"
                          --include "{field}*{t}-ir_seg.fits.gz"
                          """.replace('\n', ' '))
        
    wcs = db.SQL(f"""SELECT * FROM combined_tiles 
                     WHERE tile = '{xm:02d}.{ym:02d}'
                     AND field = '{field}'""")
    
    h = pyfits.Header()
    for c in wcs.colnames:
        if c in ['footprint']:
            continue
        
        h[c] = wcs[c][0]
    
    xnpix = h['NAXIS1']
    ynpix = h['NAXIS2']
    
    olap = 256
    pad = 32
    
    h['NAXIS1'] = h['NAXIS1']*nx - olap*(nx-1)
    h['NAXIS2'] = h['NAXIS2']*ny - olap*(ny-1)
    
    img_shape = (h['NAXIS2'], h['NAXIS1'])
    sci = np.zeros(img_shape, dtype=np.float32)
    wht = np.zeros(img_shape, dtype=np.float32)
    seg = np.zeros(img_shape, dtype=int)
    
    skip_tiles = []
    
    for tile, txi, tyi in zip(olap_tiles['tile'], tx, ty):
        _file = f'{field}-080-{txi:02d}.{tyi:02d}-{filt}_dr*_sci.fits.gz'
        _files = glob.glob(_file)
        if len(_files) == 0:
            msg = f'*ERROR* {_file} not found'
            utils.log_comment(utils.LOGFILE, msg, verbose=True)
            continue
            
        file = _files[0]
        
        msg = f'Add tile {file} to {assoc} mosaic'
        utils.log_comment(utils.LOGFILE, msg, verbose=True)

        im = pyfits.open(file)
        if im[0].shape[0] != xnpix:
            print(f'Bad size ({im[0].data.shape})')
            skip_tiles.append(tile)
            continue
            
        olap_dx = (txi-xm)*olap
        olap_dy = (tyi-ym)*olap
        
        slx = slice((txi-xm)*xnpix - olap_dx + pad,
                    (txi-xm+1)*xnpix - olap_dx)
        sly = slice((tyi-ym)*ynpix - olap_dy + pad,
                    (tyi-ym+1)*ynpix - olap_dy)
        
        for k in im[0].header:
            if k not in h:
                h[k] = im[0].header[k]
                
        sci[sly, slx] = im[0].data[pad:, pad:]*1

        im = pyfits.open(file.replace('_sci','_wht'))
        wht[sly, slx] = im[0].data[pad:, pad:]**1

        im = pyfits.open(file.replace('_drz_sci','_seg'))
        seg[sly, slx] = im[0].data[pad:, pad:]**1
    
    _hdu = pyfits.PrimaryHDU(data=sci, header=h)
    _hdu.writeto(f'{assoc}-{filt}_drz_sci.fits', overwrite=True)
    
    _hdu = pyfits.PrimaryHDU(data=wht, header=h)
    _hdu.writeto(f'{assoc}-{filt}_drz_wht.fits', overwrite=True)
    
    _hdu = pyfits.PrimaryHDU(data=seg, header=h)
    _hdu.writeto(f'{assoc}-ir_seg.fits', overwrite=True)
    
    if clean:
        for tile, txi, tyi in zip(olap_tiles['tile'], tx, ty):
            files = glob.glob(f'cos-080-{txi:02d}.{tyi:02d}-{filt}*')
            for file in files:
                msg = f'rm {file}'
                utils.log_comment(utils.LOGFILE, msg, verbose=True)
                os.remove(file)

    #### Catalog
    cols = ['id','thresh', 'npix', 'tnpix',
            'xminpix as xmin', 'xmaxpix as xmax',
            'yminpix as ymin', 'ymaxpix as ymax', 
            'x', 'y', 'x2_image', 'y2_image', 'xy_image',
            'errx2', 'erry2', 'errxy',
            'a_image', 'b_image', 'theta_image',
            'cxx_image', 'cyy_image', 'cxy_image',
            'cflux', 'flux',
            'cpeak', 'peak', 'xcpeak', 'ycpeak', 'xpeak', 'ypeak',
            'flag',
            'x_image', 'y_image',
            #'x_world', 'y_world',
            'ra as x_world', 'dec as y_world', 
            'flux_iso', 'fluxerr_iso', 'area_iso', 'mag_iso', 'kron_radius',
            'kron_rcirc', 'flux_auto', 'fluxerr_auto', 'bkg_auto',
            'flag_auto', 'area_auto', 'flux_radius_flag', 'flux_radius_20',
            'flux_radius', 'flux_radius_90', 'tot_corr', 'mag_auto',
            'magerr_auto', 'flux_aper_0', 'fluxerr_aper_0', 'flag_aper_0',
            'bkg_aper_0', 'mask_aper_0', 'flux_aper_1', 'fluxerr_aper_1',
            'flag_aper_1', 'bkg_aper_1', 'mask_aper_1', 'flux_aper_2',
            'fluxerr_aper_2', 'flag_aper_2', 'bkg_aper_2', 'mask_aper_2',
            'flux_aper_3', 'fluxerr_aper_3', 'flag_aper_3', 'bkg_aper_3',
            'mask_aper_3']
    
    scols = ','.join(cols)
    
    tabs = []
    for tile, txi, tyi in zip(olap_tiles['tile'], tx, ty):
        if tile in skip_tiles:
            continue
            
        _SQL = f"""SELECT {scols} from combined_tile_phot
                   where tile = '{tile}'
                   AND x < 2048 AND y < 2048
                   AND x > {pad} AND y > {pad}"""
                   
        tab = db.SQL(_SQL)
                
        # Pixel offsets
        olap_dx = (txi-xm)*olap
        olap_dy = (tyi-ym)*olap

        dx = (txi-xm)*(2048+olap) - olap_dx
        dy = (tyi-ym)*(2048+olap) - olap_dx
        
        for c in tab.colnames:
            if c in ['xmin', 'xmax', 'x', 'x_image', 'xpeak', 'xcpeak']:
                tab[c] += dx
            elif c in ['ymin', 'ymax', 'y', 'y_image', 'ypeak', 'ycpeak']:
                tab[c] += dy
        
        tabs.append(tab)
        
        msg = f'{assoc}: Query {tile} catalog N={len(tab)} dx={dx} dy={dy}'
        utils.log_comment(utils.LOGFILE, msg, verbose=True)
    
    tab = astropy.table.vstack(tabs)
    for c in tab.colnames:
        tab.rename_column(c, c.upper())
    
    # Make it look as expected for grizli model
    if hasattr(tab['MAG_AUTO'], 'filled'):
        tab['MAG_AUTO'] = tab['MAG_AUTO'].filled(99)
        
    tab.rename_column('ID','NUMBER')
    tab['NUMBER'] = tab['NUMBER'].astype(int)
    tab.write(f'{assoc}-ir.cat.fits', overwrite=True)
    
    ### Grism
    if 0:
        os.system(f'aws s3 sync s3://grizli-v2/HST/Pipeline/{assoc}/Prep/ ./ --exclude "*" --include "*flt.fits" --include "*yml"')
    
        # grism_files = glob.glob('iehn*[a-p]_flt.fits')
        # grism_files.sort()
        #     
        # grp = auto_script.grism_prep(field_root=assoc, 
        #                              gris_ref_filters={'G141':['ir']},
        #                              files=grism_files,
        #                              refine_mag_limits=[18,23], 
        #                              PREP_PATH='./')
        # 
        # if len(glob.glob(f'{assoc}*_grism*fits*')) == 0:
        #     grism_files = glob.glob('*GrismFLT.fits')
        #     grism_files.sort()
        # 
        #     catalog = glob.glob(f'{assoc}-*.cat.fits')[0]
        #     try:
        #         seg_file = glob.glob(f'{assoc}-*_seg.fits')[0]
        #     except:
        #         seg_file = None
        # 
        #     grp = multifit.GroupFLT(grism_files=grism_files, direct_files=[], 
        #                             ref_file=None, seg_file=seg_file, 
        #                             catalog=catalog, cpu_count=-1, sci_extn=1, 
        #                             pad=256)
        # 
        #     # Make drizzle model images
        #     grp.drizzle_grism_models(root=assoc, kernel='point', scale=0.15)
        # 
        #     # Free grp object
        #     del(grp)
        # 
        # pline = auto_script.DITHERED_PLINE.copy()
        # args_file = f'{assoc}_fit_args.npy'
        # 
        # if (not os.path.exists(args_file)):
        # 
        #     msg = '# generate_fit_params: ' + args_file
        #     utils.log_comment(utils.LOGFILE, msg, verbose=True, show_date=True)
        # 
        #     pline['pixscale'] = 0.1 #mosaic_args['wcs_params']['pixel_scale']
        #     pline['pixfrac'] = 0.5  #mosaic_args['mosaic_pixfrac']
        #     if pline['pixfrac'] > 0:
        #         pline['kernel'] = 'square'
        #     else:
        #         pline['kernel'] = 'point'
        # 
        #     min_sens = 1.e-4
        #     min_mask = 1.e-4
        # 
        #     fit_trace_shift = True
        # 
        #     args = auto_script.generate_fit_params(field_root=assoc, prior=None, MW_EBV=0.0, pline=pline, fit_only_beams=True, run_fit=True, poly_order=7, fsps=True, min_sens=min_sens, min_mask=min_mask, sys_err=0.03, fcontam=0.2, zr=[0.05, 3.4], save_file=args_file, fit_trace_shift=fit_trace_shift, include_photometry=False, use_phot_obj=False)
        # 
        #     os.system(f'cp {args_file} fit_args.npy')

    
